import numpy as np
import os
import pydoc
from rich import print
import shutil
from typing import Union

from sfaira.commands.utils import get_pydoc
from sfaira.consts.utils import clean_doi
from sfaira.data import DatasetGroupDirectoryOriented, DatasetGroup, DatasetBase
from sfaira.data.utils import read_yaml

try:
    import sfaira_extension as sfairae
except ImportError:
    sfairae = None


class DataloaderAnnotater:

    def __init__(self):
        self.WD = os.path.dirname(__file__)
        self.file_path = None
        self.meta_path = None
        self.cache_path = None
        self.pydoc_handle = None

    def annotate(self, path_loader: str, path_data: str, doi: Union[str, None]):
        """
        Annotates a provided dataloader.

        Moderate the suggestions made here: Choose the best fit cell ontology label for your cells.
        Sfaira uses multiple mechanisms of finding matches, depending on how the free text was generated, these might be
        differentially successful. The proposed IDs groups are separate by ":|||:" strings to give you a visual anchor
        when going through these lists. You need to delete all of these division strings and all labels in the second
        columns other than the best fit label. Do not change the first column,
        (Note that columns are separated by ",")
        You can also manually check maps here: https://www.ebi.ac.uk/ols/ontologies/cl
        """
        doi_sfaira_repr = clean_doi(doi)
        self._setup_loader(path_loader, doi_sfaira_repr)
        self._annotate(path_data, path_loader, doi, doi_sfaira_repr)

    def _setup_loader(self, path_loader: str, doi_sfaira_repr: str):
        """
        Define the file names, loader paths and base paths of loader collections for sfaira and sfaira_extension
        """
        file_path, pydoc_handle = get_pydoc(path_loader=path_loader, doi_sfaira_repr=doi_sfaira_repr)
        meta_path = None
        cache_path = None
        # Clear dataset cache
        shutil.rmtree(cache_path, ignore_errors=True)

        self.file_path = file_path
        self.meta_path = meta_path
        self.cache_path = cache_path
        self.pydoc_handle = pydoc_handle

    def _get_ds(self, test_data: str):
        ds = DatasetGroupDirectoryOriented(
            file_base=self.file_path,
            data_path=test_data,
            meta_path=None,
            cache_path=None
        )
        return ds

    def buffered_load(self, test_data: str, doi_sfaira_repr: str):
        if not os.path.exists(test_data):
            raise ValueError(f"test-data directory {test_data} does not exist.")
        if doi_sfaira_repr not in os.listdir(test_data):
            raise ValueError(f"did not find data folder named {doi_sfaira_repr} in test-data directory "
                             f"{test_data}, only found {os.listdir(test_data)}")
        ds = self._get_ds(test_data=test_data)
        ds.load(
            load_raw=True,  # Force raw load so non confound future tests by data loader bugs in previous versions.
            allow_caching=False,
            verbose=3
        )
        assert len(ds.ids) > 0, f"no data sets loaded, make sure raw data is in {test_data}, "\
                                f"found {os.listdir(os.path.join(test_data, doi_sfaira_repr))}"
        return ds

    def _annotate(self, test_data: str, path: str, doi: str, doi_sfaira_repr: str):
        ds = self.buffered_load(test_data=test_data, doi_sfaira_repr=doi_sfaira_repr)
        # Create cell type conversion table:
        cwd = os.path.dirname(self.file_path)
        # Group data sets by file module:
        # Note that if we were not grouping the cell type map .tsv files by file module, we could directly call
        # write_ontology_class_map on the ds.
        tsvs_written = []
        for f in os.listdir(cwd):
            if os.path.isfile(os.path.join(cwd, f)):  # only files
                # Narrow down to data set files:
                if f.split(".")[-1] == "py" and f.split(".")[0] not in ["__init__", "base", "group"]:
                    file_module = ".".join(f.split(".")[:-1])

                    # I) Instantiate Data set group to get all IDs of data sets associated with this .py file.
                    # Note that all data sets in this directory are already loaded in ds, so we just need the IDs.
                    DatasetFound = pydoc.locate(self.pydoc_handle + "." + file_module + ".Dataset")
                    # Load objects from name space:
                    # - load(): Loading function that return anndata instance.
                    # - SAMPLE_FNS: File name list for DatasetBaseGroupLoadingManyFiles
                    load_func = pydoc.locate(self.pydoc_handle + "." + file_module + ".load")
                    load_func_annotation = pydoc.locate(self.pydoc_handle + "." + file_module + ".LOAD_ANNOTATION")
                    sample_fns = pydoc.locate(self.pydoc_handle + "." + file_module + ".SAMPLE_FNS")
                    fn_yaml = os.path.join(cwd, file_module + ".yaml")
                    fn_yaml = fn_yaml if os.path.exists(fn_yaml) else None
                    # Check for sample_fns in yaml:
                    if fn_yaml is not None:
                        assert os.path.exists(fn_yaml), f"did not find yaml {fn_yaml}"
                        yaml_vals = read_yaml(fn=fn_yaml)
                        if sample_fns is None and yaml_vals["meta"]["sample_fns"] is not None:
                            sample_fns = yaml_vals["meta"]["sample_fns"]
                    if sample_fns is None:
                        sample_fns = [None]
                    # Here we distinguish between class that are already defined and those that are not.
                    # The latter case arises if meta data are defined in YAMLs and _load is given as a function.
                    if DatasetFound is None:
                        datasets_f = [
                            DatasetBase(
                                data_path=test_data,
                                meta_path=self.meta_path,
                                cache_path=self.cache_path,
                                load_func=load_func,
                                dict_load_func_annotation=load_func_annotation,
                                sample_fn=x,
                                sample_fns=sample_fns if sample_fns != [None] else None,
                                yaml_path=fn_yaml,
                            ) for x in sample_fns
                        ]
                    else:
                        datasets_f = [
                            DatasetFound(
                                data_path=test_data,
                                meta_path=self.meta_path,
                                cache_path=self.cache_path,
                                load_func=load_func,
                                load_func_annotation=load_func_annotation,
                                sample_fn=x,
                                sample_fns=sample_fns if sample_fns != [None] else None,
                                yaml_path=fn_yaml,
                            ) for x in sample_fns
                        ]
                    # II) Build a data set group from the already loaded data sets and use the group ontology writing
                    # function.
                    dsg_f = DatasetGroup(datasets=dict([(x.id, ds.datasets[x.id]) for x in datasets_f]))
                    # III) Write this directly into the sfaira clone so that it can be committed via git.
                    doi_sfaira_repr = clean_doi(doi)
                    fn_tsv = os.path.join(path, doi_sfaira_repr, f"{file_module}")
                    # Define .tsvs to write:
                    attrs = [
                        k for k in dsg_f._adata_ids.ontology_constrained
                        if np.any([getattr(v, k + "_obs_key") is not None for v in dsg_f.datasets.values()])
                    ]
                    dsg_f.write_ontology_class_maps(
                        fn=fn_tsv,
                        attrs=attrs,
                        protected_writing=True,
                        n_suggest=4,
                    )
                    tsvs_written.append((fn_tsv, attrs))
        print("[bold blue]Completed annotation.")
        print('[bold orange]Sfaira butler: "Up next, follow these steps until the next call of the sfaira CLI:"')
        self.action_counter = 1
        print(f'[bold orange]               "{self.action_counter}) Proceed to chose ontology symbols for each free '
              f'text label in the tsv files:"')
        for prefix, attrs in tsvs_written:
            for attr in attrs:
                print(f'[bold orange]                    -{prefix}_{attr}.tsv')
        print('[bold orange]                "Each tsv has two columns: free text labels found in the data on the left '
              'and suggestions on the right."')
        print('[bold orange]                "Each suggested symbol lies between two : characters."')
        print('[bold orange]                ": is a separator between suggested symbols and :|||: between symbol '
              'groups that were found through different search strategies."')
        print('[bold orange]                "Take care to not remove the tab separators in the table."')
        print('[bold orange]                "You only need to finish the second column now."')
        print('[bold orange]                "The third column with ontology IDs is added by `finalize` in phase 3."')
        self.action_counter += 1
        print(f'[bold orange]               "{self.action_counter}) Then proceed to finish .yaml file if not already '
              f'done."')
        self.action_counter += 1
        print(f'[bold orange]               "{self.action_counter}) Then proceed to phase 3 '
              f'sfaira finalize-dataloader."')
        self.action_counter += 1



import os
import shutil
from typing import Union

from sfaira import settings


def clean_cache(cache: Union[None, str] = None):
    """
    Utility function to clean cached objects in paths of sfaira installation.

    This can be used to force re-caching or to reduce directory size.
    """
    if cache is not None:
        cache_dir_dict = {
            "all": settings.cachedir_base,
            "dataset_meta": settings.cachedir_databases,
            "genomes": settings.cachedir_genomes,
            "ontologies": settings.cachedir_ontologies,
        }
        if cache not in cache_dir_dict.keys():
            raise ValueError(f"Did not find cache directory input {cache} in support list: "
                             f"{list(cache_dir_dict.keys())}")
        else:
            print(f"cleaning cache {cache} in directory {cache_dir_dict[cache]}")
            # Assert that a path within sfaira is selected as a sanity check:
            dir_to_delete = cache_dir_dict[cache]
            dir_sfaira = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
            assert str(dir_to_delete).startswith(dir_sfaira), \
                f"trying to delete outside of sfaira installation: {dir_to_delete}"
            shutil.rmtree(dir_to_delete)


def clean_doi(doi: str):
    # Check if DOI is already in sfaira format, in that case do not prefix with "d":
    if doi.startswith("d10_") or doi.startswith("dno_doi_"):
        return f'{doi.translate({ord(c): "_" for c in r"!@#$%^&*()[]/{};:,.<>?|`~-=_+"})}'
    else:
        return f'd{doi.translate({ord(c): "_" for c in r"!@#$%^&*()[]/{};:,.<>?|`~-=_+"})}'


def clean_id_str(s):
    if s is not None:
        s = s.replace(',', '').replace(' ', '').replace('-', '').replace('_', '').replace("'", '').lower()
    return s



import anndata
import numpy as np
import os
import pandas as pd
import scipy.sparse
import zipfile

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://ndownloader.figshare.com/files/17727365"
        self.download_url_meta = [
            "https://ndownloader.figshare.com/files/21758835",
            "https://ndownloader.figshare.com/files/22447898",
        ]

        self.author = "Han"
        self.doi_journal = "10.1038/s41586-020-2157-4"
        self.healthy = True
        self.layer_counts = "X"
        self.organism = "Homo sapiens"
        self.assay_sc = "microwell-seq"
        self.state_exact = "healthy"
        self.year = 2020
        self.sample_source = "primary_tissue"

        self.bio_sample_obs_key = "sample"
        self.cell_type_obs_key = "celltype_specific"
        self.development_stage_obs_key = "dev_stage"
        self.organ_obs_key = "organ"
        self.primary_data = True
        self.sex_obs_key = "sex"
        self.age_obs_key = "age"

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    sample_organ_dict = {
        'AdultAdipose_1': 'adipose tissue of abdominal region',
        'AdultAdrenalGland_2': 'adrenal gland',
        'AdultAdrenalGland_3': 'adrenal gland',
        'AdultArtery_1': 'artery',
        'AdultAscendingColon_1': 'ascending colon',
        'AdultBladder_1': 'urinary bladder',
        'AdultBladder_2': 'urinary bladder',
        'AdultCerebellum_1': 'cerebellum',
        'AdultCervix_1': 'uterine cervix',
        'AdultColon_1': 'colon',
        'AdultDuodenum_1': 'duodenum',
        'AdultEpityphlon_1': 'caecum',
        'AdultEsophagus_1': 'esophagus',
        'AdultEsophagus_2': 'esophagus',
        'AdultFallopiantube_1': 'fallopian tube',
        'AdultGallbladder_1': 'gall bladder',
        'AdultGallbladder_2': 'gall bladder',
        'AdultHeart_1': 'heart',
        'AdultHeart_2': 'heart',
        'AdultIleum_2': 'ileum',
        'AdultJejunum_2': 'jejunum',
        'AdultKidney_2': 'kidney',
        'AdultKidney_3': 'kidney',
        'AdultKidney_4': 'kidney',
        'AdultLiver_1': 'liver',
        'AdultLiver_2': 'liver',
        'AdultLiver_4': 'liver',
        'AdultLung_1': 'lung',
        'AdultLung_2': 'lung',
        'AdultLung_3': 'lung',
        'AdultMuscle_1': 'skeletal muscle organ',
        'AdultOmentum_1': 'omentum',
        'AdultOmentum_2': 'omentum',
        'AdultOmentum_3': 'omentum',
        'AdultPancreas_1': 'pancreas',
        'AdultPeripheralBlood_3': 'blood',
        'AdultPeripheralBlood_4': 'blood',
        'AdultPleura_1': 'pleura',
        'AdultProstate_1': 'prostate gland',
        'AdultRectum_1': 'rectum',
        'AdultSigmoidColon_1': 'sigmoid colon',
        'AdultSpleenParenchyma_1': 'parenchyma of spleen',
        'AdultSpleen_1': 'spleen',
        'AdultStomach_1': 'stomach',
        'AdultStomach_2': 'stomach',
        'AdultStomach_3': 'stomach',
        'AdultTemporalLobe_1': 'temporal lobe',
        'AdultThyroid_1': 'thyroid gland',
        'AdultThyroid_2': 'thyroid gland',
        'AdultTrachea_2': 'trachea',
        'AdultTransverseColon_2': 'transverse colon',
        'AdultUreter_1': 'ureter',
        'AdultUterus_1': 'uterus',
        'BoneMarrow_1': 'bone marrow',
        'BoneMarrow_2': 'bone marrow',
        'ChorionicVillus_1': 'chorionic villus',
        'CordBloodCD34P_1': 'umbilical cord blood',
        'CordBloodCD34P_2': 'umbilical cord blood',
        'CordBlood_1': 'umbilical cord blood',
        'CordBlood_2': 'umbilical cord blood',
        'FetalAdrenalGland_2': 'adrenal gland',
        'FetalAdrenalGland_3': 'adrenal gland',
        'FetalAdrenalGland_4': 'adrenal gland',
        'FetalBrain_3': 'brain',
        'FetalBrain_4': 'brain',
        'FetalBrain_5': 'brain',
        'FetalBrain_6': 'brain',
        'FetalCalvaria_1': 'vault of skull',
        'FetalEyes_1': 'eye',
        'FetalFemaleGonad_1': 'ovary',
        'FetalFemaleGonad_2': 'ovary',
        'FetalHeart_1': 'heart',
        'FetalHeart_2': 'heart',
        'FetalIntestine_1': 'intestine',
        'FetalIntestine_2': 'intestine',
        'FetalIntetsine_3': 'intestine',
        'FetalIntestine_4': 'intestine',
        'FetalIntestine_5': 'intestine',
        'FetalKidney_3': 'kidney',
        'FetalKidney_4': 'kidney',
        'FetalKidney_5': 'kidney',
        'FetalKidney_6': 'kidney',
        'FetalLung_1': 'lung',
        'FetalLung_2': 'lung',
        'FetalMaleGonad_1': 'testis',
        'FetalMaleGonad_2': 'testis',
        'FetalMuscle_1': 'skeletal muscle organ',
        'FetalPancreas_1': 'pancreas',
        'FetalPancreas_2': 'pancreas',
        'FetalPancreas_3': 'pancreas',
        'FetalRib_2': 'rib',
        'FetalRib_3': 'rib',
        'FetalSkin_2': 'skin of body',
        'FetalSkin_3': 'skin of body',
        'FetalSpinalCord_1': 'spinal cord',
        'FetalStomach_1': 'stomach',
        'FetalStomach_2': 'stomach',
        'FetalThymus_1': 'thymus',
        'FetalThymus_2': 'thymus',
        'HESC_1': 'blastocyst',
        'Liver_1': 'liver',
        'Liver_2': 'liver',
        'NeonatalAdrenalGland_1': 'adrenal gland',
        'PeripheralBlood_1': 'blood',
        'Placenta_1': 'placenta',
    }
    sample_dev_stage_dict = {
        'AdultAdipose_1': 'homosapiens adult stage',
        'AdultAdrenalGland_2': 'homosapiens adult stage',
        'AdultAdrenalGland_3': 'homosapiens adult stage',
        'AdultArtery_1': 'homosapiens adult stage',
        'AdultAscendingColon_1': 'homosapiens adult stage',
        'AdultBladder_1': 'homosapiens adult stage',
        'AdultBladder_2': 'homosapiens adult stage',
        'AdultCerebellum_1': 'homosapiens adult stage',
        'AdultCervix_1': 'homosapiens adult stage',
        'AdultColon_1': 'homosapiens adult stage',
        'AdultDuodenum_1': 'homosapiens adult stage',
        'AdultEpityphlon_1': 'homosapiens adult stage',
        'AdultEsophagus_1': 'homosapiens adult stage',
        'AdultEsophagus_2': 'homosapiens adult stage',
        'AdultFallopiantube_1': 'homosapiens adult stage',
        'AdultGallbladder_1': 'homosapiens adult stage',
        'AdultGallbladder_2': 'homosapiens adult stage',
        'AdultHeart_1': 'homosapiens adult stage',
        'AdultHeart_2': 'homosapiens adult stage',
        'AdultIleum_2': 'homosapiens adult stage',
        'AdultJejunum_2': 'homosapiens adult stage',
        'AdultKidney_2': 'homosapiens adult stage',
        'AdultKidney_3': 'homosapiens adult stage',
        'AdultKidney_4': 'homosapiens adult stage',
        'AdultLiver_1': 'homosapiens adult stage',
        'AdultLiver_2': 'homosapiens adult stage',
        'AdultLiver_4': 'homosapiens adult stage',
        'AdultLung_1': 'homosapiens adult stage',
        'AdultLung_2': 'homosapiens adult stage',
        'AdultLung_3': 'homosapiens adult stage',
        'AdultMuscle_1': 'homosapiens adult stage',
        'AdultOmentum_1': 'homosapiens adult stage',
        'AdultOmentum_2': 'homosapiens adult stage',
        'AdultOmentum_3': 'homosapiens adult stage',
        'AdultPancreas_1': 'homosapiens adult stage',
        'AdultPeripheralBlood_3': 'homosapiens adult stage',
        'AdultPeripheralBlood_4': 'homosapiens adult stage',
        'AdultPleura_1': 'homosapiens adult stage',
        'AdultProstate_1': 'homosapiens adult stage',
        'AdultRectum_1': 'homosapiens adult stage',
        'AdultSigmoidColon_1': 'homosapiens adult stage',
        'AdultSpleenParenchyma_1': 'homosapiens adult stage',
        'AdultSpleen_1': 'homosapiens adult stage',
        'AdultStomach_1': 'homosapiens adult stage',
        'AdultStomach_2': 'homosapiens adult stage',
        'AdultStomach_3': 'homosapiens adult stage',
        'AdultTemporalLobe_1': 'homosapiens adult stage',
        'AdultThyroid_1': 'homosapiens adult stage',
        'AdultThyroid_2': 'homosapiens adult stage',
        'AdultTrachea_2': 'homosapiens adult stage',
        'AdultTransverseColon_2': 'homosapiens adult stage',
        'AdultUreter_1': 'homosapiens adult stage',
        'AdultUterus_1': 'homosapiens adult stage',
        'BoneMarrow_1': 'homosapiens adult stage',
        'BoneMarrow_2': 'homosapiens adult stage',
        'ChorionicVillus_1': 'homosapiens adult stage',
        'CordBloodCD34P_1': 'homosapiens adult stage',
        'CordBloodCD34P_2': 'homosapiens adult stage',
        'CordBlood_1': 'homosapiens adult stage',
        'CordBlood_2': 'homosapiens adult stage',
        'FetalAdrenalGland_2': 'fetal stage',
        'FetalAdrenalGland_3': 'fetal stage',
        'FetalAdrenalGland_4': 'fetal stage',
        'FetalBrain_3': 'fetal stage',
        'FetalBrain_4': 'fetal stage',
        'FetalBrain_5': 'fetal stage',
        'FetalBrain_6': 'fetal stage',
        'FetalCalvaria_1': 'fetal stage',
        'FetalEyes_1': 'fetal stage',
        'FetalFemaleGonad_1': 'fetal stage',
        'FetalFemaleGonad_2': 'fetal stage',
        'FetalHeart_1': 'fetal stage',
        'FetalHeart_2': 'fetal stage',
        'FetalIntestine_1': 'fetal stage',
        'FetalIntestine_2': 'fetal stage',
        'FetalIntetsine_3': 'fetal stage',
        'FetalIntestine_4': 'fetal stage',
        'FetalIntestine_5': 'fetal stage',
        'FetalKidney_3': 'fetal stage',
        'FetalKidney_4': 'fetal stage',
        'FetalKidney_5': 'fetal stage',
        'FetalKidney_6': 'fetal stage',
        'FetalLung_1': 'fetal stage',
        'FetalLung_2': 'fetal stage',
        'FetalMaleGonad_1': 'fetal stage',
        'FetalMaleGonad_2': 'fetal stage',
        'FetalMuscle_1': 'fetal stage',
        'FetalPancreas_1': 'fetal stage',
        'FetalPancreas_2': 'fetal stage',
        'FetalPancreas_3': 'fetal stage',
        'FetalRib_2': 'fetal stage',
        'FetalRib_3': 'fetal stage',
        'FetalSkin_2': 'fetal stage',
        'FetalSkin_3': 'fetal stage',
        'FetalSpinalCord_1': 'fetal stage',
        'FetalStomach_1': 'fetal stage',
        'FetalStomach_2': 'fetal stage',
        'FetalThymus_1': 'fetal stage',
        'FetalThymus_2': 'fetal stage',
        'HESC_1': 'blastula stage',
        'Liver_1': 'homosapiens adult stage',
        'Liver_2': 'homosapiens adult stage',
        'NeonatalAdrenalGland_1': 'newborn homosapiens stage',
        'PeripheralBlood_1': 'homosapiens adult stage',
        'Placenta_1': 'homosapiens adult stage',
    }
    sex_dict = {
        'Male': "male",
        'Female': "female",
        'nan': "unknown",
        'FeM=male': "unknown",
    }

    adata = anndata.read(os.path.join(data_dir, "HCL_Fig1_adata.h5ad"))
    # convert to sparse matrix
    adata.X = scipy.sparse.csr_matrix(adata.X).copy()

    # harmonise annotations
    for col in ["batch", "tissue"]:
        adata.obs[col] = adata.obs[col].astype("str")
    adata.obs.index = adata.obs.index.str.replace("AdultJeJunum", "AdultJejunum", regex=True).str.replace(
        "AdultGallBladder", "AdultGallbladder", regex=True).str.replace(
        "FetalFemaleGonald", "FetalFemaleGonad", regex=True)
    adata.obs.replace({"AdultJeJunum": "AdultJejunum", "AdultGallBladder": "AdultGallbladder",
                       "FetalFemaleGonald": "FetalFemaleGonad"}, regex=True, inplace=True)
    adata.obs.index = ["-".join(i.split("-")[:-1]) for i in adata.obs.index]

    # load celltype labels and harmonise them
    # This pandas code should work with pandas 1.2 but it does not and yields an empty data frame:
    fig1_anno = pd.read_excel(
        os.path.join(data_dir, "HCL_Fig1_cell_Info.xlsx"),
        index_col="cellnames",
        engine="xlrd",  # ToDo: Update when pandas xlsx reading with openpyxl is fixed: yields empty tables
    )
    fig1_anno.index = fig1_anno.index.str.replace("AdultJeJunum", "AdultJejunum", regex=True).str.replace(
        "AdultGallBladder", "AdultGallbladder", regex=True).str.replace(
        "FetalFemaleGonald", "FetalFemaleGonad", regex=True)

    # check that the order of cells and cell labels is the same
    assert np.all(fig1_anno.index == adata.obs.index)

    # add annotations to adata object and rename columns
    adata.obs = pd.concat([adata.obs, fig1_anno[["cluster", "stage", "donor", "celltype"]]], axis=1)
    adata.obs.columns = ["sample", "tissue", "n_genes", "n_counts", "cluster_global", "stage", "donor",
                         "celltype_global"]

    # add sample-wise annotations to the full adata object
    df = pd.DataFrame(
        columns=["Cell_barcode", "Sample", "Batch", "Cell_id", "Cluster_id", "Ages", "Development_stage", "Method",
                 "Gender", "Source", "Biomaterial", "Name", "ident", "Celltype"])
    archive = zipfile.ZipFile(os.path.join(data_dir, "annotation_rmbatch_data_revised417.zip"))
    for f in archive.namelist():
        df1 = pd.read_csv(archive.open(f), encoding="unicode_escape")
        df = pd.concat([df, df1], sort=True)
    df = df.set_index("Cell_id")
    adata = adata[[i in df.index for i in adata.obs.index]].copy()
    a_idx = adata.obs.index.copy()
    adata.obs = pd.concat([adata.obs, df[
        ["Ages", "Celltype", "Cluster_id", "Gender", "Method", "Source"]
    ]], axis=1)
    assert np.all(a_idx == adata.obs.index)

    # remove musmusculus cells from the object  # ToDo: add this back in as musmusculus data sets?
    adata = adata[adata.obs["Source"] != "MCA2.0"].copy()

    # tidy up the column names of the obs annotations
    adata.obs.columns = [
        "sample", "sub_tissue", "n_genes", "n_counts", "cluster_global", "dev_stage", "donor", "celltype_global",
        "age", "celltype_specific", "cluster_specific", "sex", "assay_sc", "source"]
    # Remove new line characters from cell type:
    adata.obs["celltype_specific"] = [
        x.replace("\n", "").rstrip()
        for x in adata.obs["celltype_specific"].values
    ]
    adata.obs["organ"] = [sample_organ_dict[x] for x in adata.obs["sample"].values]
    adata.obs["sex"] = [sex_dict[str(x)] for x in adata.obs["sex"].values]
    # TODO are the more exact developmental stages in dev_stage?
    adata.obs["dev_stage"] = [sample_dev_stage_dict[str(x)] for x in adata.obs["sample"].values]

    return adata



import numpy as np

from sfaira.data import load_store
from sfaira.versions.genomes.genomes import GenomeContainer

from sfaira.unit_tests.data_for_tests.loaders import RELEASE_MOUSE, PrepareData


def _get_cart(store_format, feature_space, map_fn=None, batch_size=1, **kwargs):
    """
    TODO move into PrepareData?
    """
    store_path = PrepareData().prepare_store(store_format=store_format)
    store = load_store(cache_path=store_path, store_format=store_format)
    store.subset(attr_key="organism", values=["Mus musculus"])
    if feature_space == "single":
        store = store.stores["Mus musculus"]
        if "idx" in kwargs.keys() and isinstance(kwargs["idx"], dict):
            assert "Mus musculus" in kwargs["idx"].keys()
            kwargs["idx"] = kwargs["idx"]["Mus musculus"]
    gc = GenomeContainer(release=RELEASE_MOUSE, organism="Mus musculus")
    gc.set(**{"biotype": "protein_coding"})
    store.genome_container = gc

    if map_fn is None:

        def map_fn(x_, obs_):
            return (np.asarray(x_), ),

    g = store.checkout(map_fn=map_fn, batch_size=batch_size, **kwargs)
    return g



import anndata
import numpy as np
import pandas as pd
import pytest
import scipy.sparse
from typing import List

from sfaira.consts import AdataIdsSfaira
from sfaira.unit_tests.tests_by_submodule.data.store.utils import _get_cart


@pytest.mark.parametrize("feature_space", ["single", "multi"])
@pytest.mark.parametrize("store_format", ["h5ad", "dao"])
def test_properties(store_format: str, feature_space: str):
    """Tests if properties of carts are available and of the right format."""
    idx = np.arange(0, 5)
    kwargs = {"idx": {"Mus musculus": idx}, "obs_keys": ["cell_type"], "randomized_batch_access": False,
              "return_dense": False}
    cart = _get_cart(store_format=store_format, feature_space=feature_space, **kwargs)

    # Meta data:
    assert cart.n_obs > len(idx)
    assert cart.n_obs_selected == len(idx)
    obs_idx = cart.obs_idx
    if feature_space != "single":
        obs_idx = obs_idx["Mus musculus"]
    if store_format == "dao":
        cart.move_to_memory()
    assert np.all(obs_idx == idx)
    # Data arrays:
    arr = [
        cart.adata,
        cart.x,
        cart.obs,
        cart.var,
    ]
    if feature_space != "single":
        arr = [x["Mus musculus"] for x in arr]
    assert np.all([x.shape[0] == len(idx) for x in arr[:3]])
    assert isinstance(arr[0], anndata.AnnData)
    assert scipy.sparse.issparse(arr[0].X)
    assert scipy.sparse.issparse(arr[1])
    assert isinstance(arr[2], pd.DataFrame)
    assert isinstance(arr[3], pd.DataFrame)
    # Other
    _ = cart.n_var


@pytest.mark.parametrize("store_format", ["h5ad", "dao"])
@pytest.mark.parametrize("idx", [np.arange(1, 10),
                                 np.concatenate([np.arange(30, 50), np.array([1, 4, 98])])])
@pytest.mark.parametrize("obs_keys", [["cell_type"]])
@pytest.mark.parametrize("randomized_batch_access", [True, False])
def test_data(store_format: str, idx, obs_keys: List[str], randomized_batch_access: bool):
    """
    Test generators queries do not throw errors and that output shapes are correct.
    """
    # Need to re-write because specific obs_keys are required:
    kwargs = {"idx": {"Mus musculus": idx}, "obs_keys": obs_keys, "randomized_batch_access": randomized_batch_access}
    cart = _get_cart(store_format=store_format, feature_space="multi", **kwargs)
    it = cart.iterator
    batch_sizes = []
    x = None
    counter = 0
    for i, z in enumerate(it()):
        counter += 1
        x_i, = z[0]
        if len(x_i.shape) == 1:
            x_i = np.expand_dims(x_i, axis=0)
        if i == 0:
            x = x_i
        batch_sizes.append(x_i.shape[0])
    assert counter > 0
    assert x.shape[1] == cart.n_var["Mus musculus"], (x.shape, cart.n_var)
    assert np.sum(batch_sizes) == len(idx), (batch_sizes, len(idx))


@pytest.mark.parametrize("store_format", ["h5ad", "dao"])
@pytest.mark.parametrize("idx", [None, np.array([1, 4, 98])])
def test_schedule_blocked(store_format: str, idx):
    """
    Test if schedule "blocks" works.
    """
    block_col = AdataIdsSfaira().cell_type
    obs_keys = [block_col]

    def map_fn(x, obs):
        return (obs, ),

    kwargs = {"idx": {"Mus musculus": idx}, "obs_keys": obs_keys, "randomized_batch_access": False,
              "batch_schedule": "blocks", "grouping": block_col, "batch_size": 0}
    cart = _get_cart(store_format=store_format, feature_space="single", map_fn=map_fn, **kwargs)
    it = cart.iterator
    block_vals = cart.obs[block_col].values
    batches = []
    for i, z in enumerate(it()):
        obs_i, = z[0]
        # Check that this batch is one single block:
        assert len(np.unique(obs_i[block_col].values)) == 1, obs_i
        batches.append(obs_i)
    assert len(batches) > 0
    # Check that one batch was emitted per block:
    assert len(np.unique(block_vals)) == len(batches), batches
    # Check that the total number of observations across blocks is correct:
    if idx is None:
        target_nobs = cart.n_obs
    else:
        target_nobs = len(idx)
    assert np.sum([y.shape[0] for y in batches]) == target_nobs


@pytest.mark.parametrize(
    "adaptor", ["python", "tensorflow", "torch", "torch-loader", "torch-iter", "torch-iter-loader"]
)
@pytest.mark.parametrize("shuffle_buffer_size", [0, 1000])
def test_adaptors(adaptor: str, shuffle_buffer_size: int):
    """
    Test if framework-specific generator adpators yield batches.
    """
    idx = np.arange(0, 10)

    def map_fn(x_, obs_):
        """
        Note: Need to convert to numpy in output because torch does not accept dask.
        """
        return (np.asarray(x_[:, :2]),),

    kwargs = {"idx": {"Mus musculus": idx}, "obs_keys": [], "randomized_batch_access": False, "retrieval_batch_size": 2,
              "map_fn": map_fn}
    cart = _get_cart(store_format="dao", feature_space="single", **kwargs)

    if adaptor == "python":
        kwargs = {}
    elif adaptor == "tensorflow":
        import tensorflow as tf

        kwargs = {"output_signature": (
            tf.TensorSpec(shape=(2,), dtype=tf.float32),
        )}
    elif adaptor in ["torch", "torch-loader", "torch-iter-loader", "torch-iter"]:
        kwargs = {}
    else:
        assert False
    it = cart.adaptor(generator_type=adaptor, shuffle_buffer=shuffle_buffer_size, **kwargs)
    if adaptor == "tensorflow":
        it = iter(it.range(2))
    if adaptor in ["torch", "torch-iter"]:
        from torch.utils.data import DataLoader
        it = list(DataLoader(it))
        it = iter(it)
    if adaptor in ["torch-loader", "torch-iter-loader"]:
        import torch
        it = iter(list(it))
    _ = next(it)



import numpy as np
import pandas as pd
from typing import Dict, List, Tuple, Union

from sfaira.consts import OCS


def synonym_string_processing(y):
    """
    Formatting of synonyms: These are richly annotated, we strip references following after either:
    BROAD, EXACT in the synonym string and characters: "'
    :param y: synonym string to process
    :return: processed synonym string
    """
    return y.lower().split("broad")[0].split("exact")[0].lower().strip("'").strip("\"").split("\" ")[0]


def _prepare_celltype_map_fuzzy(
        source,
        onto,
        onto_uberon,
        anatomical_constraint: Union[str, None],
        choices_for_perfect_match: bool,
        include_synonyms: bool,
        match_only: bool,
        n_suggest: int,
        omit_list: list,
        omit_target_list: list,
        threshold_for_partial_matching: float,
) -> Tuple[
    List[Dict[str, Union[List[str], str]]],
    List[bool]
]:
    """
    Map free text cell type names to ontology node names via fuzzy string matching and return as list.

    This function extends prepare_ontology_map_fuzzy() by additional functionalities.
    If this function does not yield good matches, consider querying this web interface:
    https://www.ebi.ac.uk/ols/index

    Search strategies:

        - exact_match: Only exact string matches to name or synonym in ontology. This is the only strategy that is
            enabled if match_only is True.
        - lenient_match: Fuzzy string matches to name or synonym in ontology based on ratio of match errors
            ((fuzz.ratio).
        - very_lenient_match: Fuzzy string matches to name or synonym in ontology based on ratio of matches
            characters from query (fuzz.partial_ratio)

    Search strategies with anatomical constraints:
    An anatomic constraint is a name of an anatomical structure that can be mapped to UBERON.

        - anatomic_onotolgy_match:
            We select cell types expected in this UBERON clade based on the link between CL and UBERON.
        - anatomic_string_match:
            We perform an additional fuzzy string matching with the anatomical structure added to the proposed
            label. This is often beneficial because analysts do not always prefix such extension (e.g. pancreatic)
            to the free text cell type labels if the entire sample consists only of cells from this anatomical
            structure. Note that if the maps from 1) were perfect, this would not be necessary. In practice, we
            find this to still recover some hits that are otherwise missed.

    Note that matches are shadowed in lower priorty strategies, ie a perfect match will not show up in the list
    of hits of any other strategy.

    :param source: Free text node labels which are to be matched to ontology nodes.
    :param onto: Cell type ontology to map into.
    :param onto_uberon: Reference UBERON ontology. Only necessary if anatomical_constraint is used.
    :param anatomical_constraint: Whether to require suggestions to be within a target anatomy defined
        within UBERON.
    :param choices_for_perfect_match: Whether to give additional matches if a perfect match was found and an
        anatomical_constraint is not not defined. This is overridden by match_only.
    :param include_synonyms: Whether to include synonyms of nodes in string search.
    :param match_only: Whether to include strict matches only in output.
    :param n_suggest: Number of cell types to suggest per search strategy.
    :param omit_list: Free text node labels to omit in map.
    :param omit_target_list: Ontology nodes to not match to.
    :param threshold_for_partial_matching: Maximum fuzzy match score below which lenient matching (ratio) is
        extended through partial_ratio.
    :return: Tuple

        - List with matches for each source, each entry is a dictionary,
            of lists of search strategies named by strategy name. If a search strategy yields perfect matches, it
            does not return a list of strings but just a single string.
        - List with boolean indicator whether or not this output should be reported.
    """
    from fuzzywuzzy import fuzz
    matches = []
    nodes = onto.nodes
    nodes = [x for x in nodes if x[1]["name"] not in omit_target_list]
    include_terms = []
    if isinstance(source, pd.DataFrame):
        source = list(zip(source.iloc[:, 0].values, source.iloc[:, 1].values))
    for x in source:
        if not isinstance(x, list) and not isinstance(x, tuple):
            x = [x, "nan"]
        term = x[0].lower().strip("'").strip("\"").strip("'").strip("\"").strip("]").strip("[")
        # Test for perfect string matching:
        scores_strict = np.array([
            np.max(
                [
                    100 if term == y[1]["name"].lower() else 0
                ] + [
                    100 if term == yy.lower() else 0
                    for yy in y[1]["synonym"]
                ]
            ) if "synonym" in y[1].keys() and include_synonyms else 100 if term == y[1]["name"].lower() else 0
            for y in nodes
        ])

        # Test for partial string matching:
        # fuzz ratio and partial_ratio capture different types of matches well, we use both here and decide below
        # which scores are used in which scenario defined through the user input.

        scores_lenient = np.array([
            np.max([fuzz.ratio(term, y[1]["name"].lower())] + [
                fuzz.ratio(term, synonym_string_processing(yy))
                for yy in y[1]["synonym"]
            ]) if "synonym" in y[1].keys() and include_synonyms else
            fuzz.ratio(term, y[1]["name"].lower())
            for y in nodes
        ])
        scores_very_lenient = np.array([
            np.max([fuzz.partial_ratio(term, y[1]["name"].lower())] + [
                fuzz.partial_ratio(term, synonym_string_processing(yy))
                for yy in y[1]["synonym"]
            ]) if "synonym" in y[1].keys() and include_synonyms else
            fuzz.partial_ratio(term, y[1]["name"].lower())
            for y in nodes
        ])
        include_terms.append(term not in omit_list)
        if match_only and not anatomical_constraint:
            # Explicitly trying to report perfect matches (match_only is True).
            matches.append({"perfect_match": [nodes[i][1]["name"] for i in np.where(scores_strict == 100)[0]][0]})
        else:
            matches_i = {}
            if np.any(scores_strict == 100) and not anatomical_constraint:
                # Perfect match and not additional information through anatomical_constraint, ie no reason to assume
                # that the user is not looking for this hit.
                matches_i.update({
                    "perfect_match": [nodes[i][1]["name"] for i in np.where(scores_strict == 100)[0]][0]
                })
                if choices_for_perfect_match:
                    matches_i.update({
                        "lenient_match": [
                            nodes[i][1]["name"] for i in np.argsort(scores_lenient)[::-1]
                            if not np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                        ][:n_suggest]
                    })
                    if np.max(scores_lenient) < threshold_for_partial_matching:
                        matches_i.update({
                            "very_lenient_match": [
                                nodes[i][1]["name"] for i in np.argsort(scores_very_lenient)[::-1]
                                if not np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                            ][:n_suggest]
                        })
            else:
                if anatomical_constraint is not None:
                    # Use anatomical constraints two fold:
                    # 1. Select cell types that are in the correct ontology.
                    # 2. Run a second string matching with the anatomical word included.

                    # 1. Select cell types that are in the correct ontology.
                    # Check that anatomical constraint is a term in UBERON and get UBERON ID:
                    anatomical_constraint_id = onto_uberon.convert_to_id(anatomical_constraint)
                    # Select up to 5 nodes which match the anatomical constraint:
                    # The entries look as follows:
                    # node.value['relationship'] = ['part_of UBERON:0001885']
                    # Find nodes that can be matched to UBERON:
                    anatomical_subselection = [
                        "relationship" in y[1].keys() and
                        np.any(["part_of UBERON" in yy for yy in y[1]["relationship"]]) and
                        np.any([
                            yy.split("part_of ")[-1] in onto_uberon.node_ids
                            for yy in y[1]["relationship"]
                        ])
                        for y in nodes
                    ]
                    uberon_ids = [
                        y[1]["relationship"][
                            np.where(["part_of UBERON" in yy for yy in y[1]["relationship"]])[0][0]
                        ].split("part_of ")[1]
                        if z else None
                        for y, z in zip(nodes, anatomical_subselection)
                    ]
                    # Check relationship in UBERON. Select for:
                    # a) parent -> a more general setting across anatomies from which one was sampled
                    # b) child -> a sub anatomy of the sampled tissue.
                    # Check this by checking if one is an ancestor of the other:
                    anatomical_subselection = [
                        z and (
                            anatomical_constraint_id in onto_uberon.get_ancestors(node=y) or
                            y in onto_uberon.get_ancestors(node=anatomical_constraint_id)
                        )
                        for y, z in zip(uberon_ids, anatomical_subselection)
                    ]
                    # Iterate over nodes sorted by string match score and masked by constraint:
                    matches_i.update({
                        "anatomic_onotolgy_match": [
                            nodes[i][1]["name"] for i in np.argsort(scores_lenient)
                            if anatomical_subselection[i] and not
                            np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                        ][-n_suggest:][::-1]
                    })

                    # 2. Run a second string matching with the anatomical word included.
                    modified_term = anatomical_constraint + " " + x[0].lower().strip("'").strip("\"").strip("]"). \
                        strip("[")
                    scores_anatomy = np.array([
                        np.max([
                            fuzz.partial_ratio(modified_term, y[1]["name"].lower())
                        ] + [
                            fuzz.partial_ratio(modified_term, synonym_string_processing(yy))
                            for yy in y[1]["synonym"]
                        ]) if "synonym" in y[1].keys() and include_synonyms else
                        fuzz.partial_ratio(modified_term, y[1]["name"].lower())
                        for y in nodes
                    ])
                    matches_i.update({
                        "anatomic_string_match": [
                            nodes[i][1]["name"] for i in np.argsort(scores_anatomy)
                            if nodes[i][1]["name"] and not
                            np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                        ][-n_suggest:][::-1]
                    })

                    # Select best overall matches based on lenient and strict matching:
                    matches_i.update({
                        "perfect_match": [nodes[i][1]["name"] for i in np.argsort(scores_strict)[::-1]][:n_suggest]
                    })
                    matches_i.update({
                        "lenient_match": [
                            nodes[i][1]["name"] for i in np.argsort(scores_lenient)[::-1]
                            if not np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                        ][:n_suggest]
                    })
                    if np.max(scores_lenient) < threshold_for_partial_matching:
                        matches_i.update({
                            "very_lenient_match": [
                                nodes[i][1]["name"] for i in np.argsort(scores_very_lenient)[::-1]
                                if not np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                            ][:n_suggest]
                        })
                else:
                    # Suggest top hits by string match:
                    matches_i.update({
                        "lenient_match": [nodes[i][1]["name"] for i in np.argsort(scores_lenient)[::-1]][:n_suggest]
                    })
                    if np.max(scores_lenient) < threshold_for_partial_matching:
                        matches_i.update({
                            "very_lenient_match": [
                                nodes[i][1]["name"] for i in np.argsort(scores_very_lenient)[::-1]
                                if not np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                            ][:n_suggest]
                        })
            matches.append(matches_i)
    return matches, include_terms


def _prepare_ontology_map_fuzzy(
        source,
        onto,
        choices_for_perfect_match: bool,
        include_synonyms: bool,
        match_only: bool,
        n_suggest: int,
        omit_list: list,
        omit_target_list: list,
        threshold_for_partial_matching: float,
) -> Tuple[
    List[Dict[str, Union[List[str], str]]],
    List[bool]
]:
    """
    Map free text node names to ontology node names via fuzzy string matching and return as list.

    If this function does not yield good matches, consider querying this web interface:
    https://www.ebi.ac.uk/ols/index

    Search strategies:

        - exact_match: Only exact string matches to name or synonym in ontology. This is the only strategy that is
            enabled if match_only is True.
        - lenient_match: Fuzzy string matches to name or synonym in ontology based on ratio of match errors
            ((fuzz.ratio).
        - very_lenient_match: Fuzzy string matches to name or synonym in ontology based on ratio of matches
            characters from query (fuzz.partial_ratio)

    Note that matches are shadowed in lower priorty strategies, ie a perfect match will not show up in the list
    of hits of any other strategy.

    :param source: Free text node labels which are to be matched to ontology nodes.
    :param onto: Ontology to map into.
    :param choices_for_perfect_match: Whether to give additional matches if a perfect match was found and an
        anatomical_constraint is not not defined. This is overridden by match_only.
    :param include_synonyms: Whether to include synonyms of nodes in string search.
    :param match_only: Whether to include strict matches only in output.
    :param n_suggest: Number of cell types to suggest per search strategy.
    :param omit_list: Free text node labels to omit in map.
    :param omit_target_list: Ontology nodes to not match to.
    :param threshold_for_partial_matching: Maximum fuzzy match score below which lenient matching (ratio) is
        extended through partial_ratio.
    :return: Tuple

        - List with matches for each source, each entry is a dictionary,
            of lists of search strategies named by strategy name. If a search strategy yields perfect matches, it
            does not return a list of strings but just a single string.
        - List with boolean indicator whether or not this output should be reported.
    """
    from fuzzywuzzy import fuzz
    source = [str(x) for x in source]
    matches = []
    nodes = onto.nodes
    nodes = [x for x in nodes if x[1]["name"] not in omit_target_list]
    include_terms = []
    if isinstance(source, pd.DataFrame):
        source = list(zip(source.iloc[:, 0].values, source.iloc[:, 1].values))
    for x in source:
        if not isinstance(x, list) and not isinstance(x, tuple):
            x = [x, "nan"]
        term = x[0].lower().strip("'").strip("\"").strip("'").strip("\"").strip("]").strip("[")
        # Test for perfect string matching:
        scores_strict = np.array([
            np.max(
                [
                    100 if term == y[1]["name"].lower() else 0
                ] + [
                    100 if term == yy.lower() else 0
                    for yy in y[1]["synonym"]
                ]
            ) if "synonym" in y[1].keys() and include_synonyms else 100 if term == y[1]["name"].lower() else 0
            for y in nodes
        ])

        # Test for partial string matching:
        # fuzz ratio and partial_ratio capture different types of matches well, we use both here and decide below
        # which scores are used in which scenario defined through the user input.

        scores_lenient = np.array([
            np.max([fuzz.ratio(term, y[1]["name"].lower())] + [
                fuzz.ratio(term, synonym_string_processing(yy))
                for yy in y[1]["synonym"]
            ]) if "synonym" in y[1].keys() and include_synonyms else
            fuzz.ratio(term, y[1]["name"].lower())
            for y in nodes
        ])
        scores_very_lenient = np.array([
            np.max([fuzz.partial_ratio(term, y[1]["name"].lower())] + [
                fuzz.partial_ratio(term, synonym_string_processing(yy))
                for yy in y[1]["synonym"]
            ]) if "synonym" in y[1].keys() and include_synonyms else
            fuzz.partial_ratio(term, y[1]["name"].lower())
            for y in nodes
        ])
        include_terms.append(term not in omit_list)
        if match_only:
            # Explicitly trying to report perfect matches (match_only is True).
            matches.append({"perfect_match": [nodes[i][1]["name"] for i in np.where(scores_strict == 100)[0]][0]})
        else:
            matches_i = {}
            if np.any(scores_strict == 100):
                # Perfect match and not additional information through anatomical_constraint, ie no reason to assume
                # that the user is not looking for this hit.
                matches_i.update({
                    "perfect_match": [nodes[i][1]["name"] for i in np.where(scores_strict == 100)[0]][0]
                })
                if choices_for_perfect_match:
                    matches_i.update({
                        "lenient_match": [
                            nodes[i][1]["name"] for i in np.argsort(scores_lenient)[::-1]
                            if not np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                        ][:n_suggest]
                    })
                    if np.max(scores_lenient) < threshold_for_partial_matching:
                        matches_i.update({
                            "very_lenient_match": [
                                nodes[i][1]["name"] for i in np.argsort(scores_very_lenient)[::-1]
                                if not np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                            ][:n_suggest]
                        })
            else:
                # Suggest top hits by string match:
                matches_i.update({
                    "lenient_match": [nodes[i][1]["name"] for i in np.argsort(scores_lenient)[::-1]][:n_suggest]
                })
                if np.max(scores_lenient) < threshold_for_partial_matching:
                    matches_i.update({
                        "very_lenient_match": [
                            nodes[i][1]["name"] for i in np.argsort(scores_very_lenient)[::-1]
                            if not np.any([nodes[i][1]["name"] in v for v in matches_i.values()])
                        ][:n_suggest]
                    })
            matches.append(matches_i)
    return matches, include_terms


def prepare_ontology_map(
        source,
        onto: str,
        anatomical_constraint: Union[str, None] = None,
        choices_for_perfect_match: bool = False,
        include_synonyms: bool = True,
        match_only: bool = False,
        n_suggest: int = 4,
        omit_list: list = [],
        omit_target_list: list = [],
        organism: Union[None, str] = None,
        threshold_for_partial_matching: float = 90.,
) -> Tuple[
    List[Dict[str, Union[List[str], str]]],
    List[bool]
]:
    """
    Map free text node names to ontology node names via fuzzy string matching and return as matching table.

    :param source: Free text node labels which are to be matched to ontology nodes.
    :param onto: Name of ontology to map to.
    :param anatomical_constraint: Whether to require suggestions to be within a target anatomy defined
        within UBERON.
    :param choices_for_perfect_match: Whether to give additional matches if a perfect match was found and an
        anatomical_constraint is not not defined. This is overridden by match_only.
    :param include_synonyms: Whether to include synonyms of nodes in string search.
    :param match_only: Whether to include strict matches only in output.
    :param n_suggest: Number of cell types to suggest per search strategy.
    :param omit_list: Free text node labels to omit in map.
    :param omit_target_list: Ontology nodes to not match to.
    :param organism: Organism of data. This is used to disambiguate the ontology in some cases, e.g. when selecting
        an ontology for the meta data item "developmental stage" for either mouse or human.
    :param threshold_for_partial_matching: Maximum fuzzy match score below which lenient matching (ratio) is
        extended through partial_ratio.
    :return: Tuple

        - List with matches for each source, each entry is a dictionary,
            of lists of search strategies named by strategy name. If a search strategy yields perfect matches, it
            does not return a list of strings but just a single string.
        - List with boolean indicator whether or not this output should be reported.
    """
    try:
        onto_cls = getattr(OCS, onto)
        if isinstance(onto_cls, dict):
            try:
                onto_cls = onto_cls[organism]
            except AttributeError as e:
                raise AttributeError(f"did not find requested organism {organism} in ontology {onto}, "
                                     f"choose from {onto_cls.keys()}.\n{e}")
    except AttributeError as e:
        raise AttributeError(f"did not find requested ontology {onto}, check spelling.\n{e}")

    if onto == "cell_type":
        fn = _prepare_celltype_map_fuzzy
        kwargs = {
            "anatomical_constraint": anatomical_constraint,
            "onto_uberon": OCS.organ,
        }
        omit_target_list = omit_target_list + ["cell"]
    else:
        fn = _prepare_ontology_map_fuzzy
        kwargs = {}
    matches, include_terms = fn(
        source=source,
        onto=onto_cls,
        choices_for_perfect_match=choices_for_perfect_match,
        include_synonyms=include_synonyms,
        match_only=match_only,
        n_suggest=n_suggest,
        omit_list=omit_list,
        omit_target_list=omit_target_list,
        threshold_for_partial_matching=threshold_for_partial_matching,
        **kwargs
    )
    return matches, include_terms


def prepare_ontology_map_tab(
        source,
        separator_suggestions: str = ":",
        separator_groups: str = ":|||:",
        **kwargs) -> pd.DataFrame:
    """
    Wraps prepare_ontology_map() to yield a table that can be saved as a .tsv.

    :param source: Free text node labels which are to be matched to ontology nodes.
    :param separator_suggestions: String separator for matches of a single strategy in output target column.
    :param separator_groups: String separator for search strategy grouped matches in output target column.
    :param kwargs: kwargs to prepare_ontology_map.
    :return: Table with source and target node names. Columns: "source", "target"
    """
    matches, include_terms = prepare_ontology_map(source=source, **kwargs)
    tab = pd.DataFrame({
        "source": source,
        "target": [
            separator_groups.join([
                separator_suggestions.join(v)
                if isinstance(v, list) else v
                for v in x.values()
            ])
            for x in matches
        ]
    })
    return tab.loc[include_terms]



import anndata
import os
import numpy as np
import scipy.sparse

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://covid19.cog.sanger.ac.uk/martin19.processed.h5ad"
        self.download_url_meta = None

        self.assay_sc = "10x 3' v2"
        self.author = "Martin"
        self.disease = "healthy"
        self.doi_journal = "10.1016/j.cell.2019.08.008"
        self.layer_counts = "X"
        self.organ = "ileum"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2019

        self.feature_id_var_key = "gene_ids"
        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"

        self.cell_type_obs_key = "CellType"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "martin19.processed.h5ad")
    adata = anndata.read(fn)
    adata.X = np.expm1(adata.X)
    adata.X = adata.X.multiply(scipy.sparse.csc_matrix(adata.obs["n_counts"].values[:, None])).multiply(1 / 10000)
    adata = adata[adata.obs["CellType"] != "Doublets"].copy()

    return adata
